"""Implements a Nvim host for python plugins."""
import functools
import imp
import inspect
import logging
import os
import os.path

from ..api import DecodeHook
from ..compat import IS_PYTHON3, find_module


__all__ = ('Host')

logger = logging.getLogger(__name__)
error, debug, info, warn = (logger.error, logger.debug, logger.info,
                            logger.warn,)


class Host(object):

    """Nvim host for python plugins.

    Takes care of loading/unloading plugins and routing msgpack-rpc
    requests/notifications to the appropriate handlers.
    """

    def __init__(self, nvim):
        """Set handlers for plugin_load/plugin_unload."""
        self.nvim = nvim
        self._specs = {}
        self._loaded = {}
        self._notification_handlers = {}
        self._request_handlers = {
            'poll': lambda: 'ok',
            'specs': lambda path: self._specs[path],
            'shutdown': self.shutdown
        }
        self._nvim_encoding = nvim.options['encoding']

    def start(self, plugins):
        """Start listening for msgpack-rpc requests and notifications."""
        self.nvim.session.run(self._on_request,
                              self._on_notification,
                              lambda: self._load(plugins))

    def shutdown(self):
        """Shutdown the host."""
        self._unload()
        self.nvim.session.stop()

    def _on_request(self, name, args):
        """Handle a msgpack-rpc request."""
        handler = self._request_handlers.get(name, None)
        if not handler:
            msg = 'no request handler registered for "%s"' % name
            warn(msg)
            raise Exception(msg)

        debug('calling request handler for "%s", args: "%s"', name, args)
        rv = handler(*args)
        debug("request handler for '%s %s' returns: %s", name, args, rv)
        return rv

    def _on_notification(self, name, args):
        """Handle a msgpack-rpc notification."""
        handler = self._notification_handlers.get(name, None)
        if not handler:
            warn('no notification handler registered for "%s"', name)
            return

        debug('calling notification handler for "%s", args: "%s"', name, args)
        handler(*args)

    def _load(self, plugins):
        for path in plugins:
            if path in self._loaded:
                raise Exception('{0} is already loaded'.format(path))
            directory, name = os.path.split(os.path.splitext(path)[0])
            file, pathname, description = find_module(name, [directory])
            try:
                module = imp.load_module(name, file, pathname, description)
            except ImportError:
                error('Encountered import error loading plugin at {0}'.format(
                    path))
            except Exception as e:
                error('Error loading plugin at {0} {1}: {2}'.format(
                    path, type(e).__name__, e))
            handlers = []
            self._discover_classes(module, handlers, path)
            self._discover_functions(module, handlers, path)
            if not handlers:
                raise Exception('{0} exports no handlers'.format(path))
            self._loaded[path] = {'handlers': handlers, 'module': module}

    def _unload(self):
        for path, plugin in self._loaded.items():
            handlers = plugin['handlers']
            for handler in handlers:
                method_name = handler._nvim_rpc_method_name
                if hasattr(handler, '_nvim_shutdown_hook'):
                    handler()
                elif handler._nvim_rpc_sync:
                    del self._request_handlers[method_name]
                else:
                    del self._notification_handlers[method_name]
        self._specs = {}
        self._loaded = {}

    def _discover_classes(self, module, handlers, plugin_path):
        for _, cls in inspect.getmembers(module, inspect.isclass):
            if getattr(cls, '_nvim_plugin', False):
                # create an instance of the plugin and pass the nvim object
                plugin = cls(self._configure_nvim_for(cls))
                # discover handlers in the plugin instance
                self._discover_functions(plugin, handlers, plugin_path)

    def _discover_functions(self, obj, handlers, plugin_path):
        predicate = lambda o: hasattr(o, '_nvim_rpc_method_name')
        specs = []
        for _, fn in inspect.getmembers(obj, predicate):
            if fn._nvim_bind:
                # bind a nvim instance to the handler
                fn2 = functools.partial(fn, self._configure_nvim_for(fn))
                # copy _nvim_* attributes from the original function
                for attr in dir(fn):
                    if attr.startswith('_nvim_'):
                        setattr(fn2, attr, getattr(fn, attr))
                fn = fn2
            # register in the rpc handler dict
            method = fn._nvim_rpc_method_name
            if fn._nvim_prefix_plugin_path:
                method = '{0}:{1}'.format(plugin_path, method)
            if fn._nvim_rpc_sync:
                if method in self._request_handlers:
                    raise Exception(('Request handler for "{0}" is ' +
                                    'already registered').format(method))
                self._request_handlers[method] = fn
            else:
                if method in self._notification_handlers:
                    raise Exception(('Notification handler for "{0}" is ' +
                                    'already registered').format(method))
                self._notification_handlers[method] = fn
            if hasattr(fn, 'nvim_rpc_spec'):
                specs.append(fn.nvim_rpc_spec)
            handlers.append(fn)
        if specs:
            self._specs[plugin_path] = specs

    def _configure_nvim_for(self, obj):
        # Configure a nvim instance for obj(checks encoding configuration)
        nvim = self.nvim
        encoding = getattr(obj, '_nvim_encoding', None)
        if IS_PYTHON3 and encoding is None:
            encoding = True
        if encoding is True:
            encoding = self._nvim_encoding
        if encoding:
            nvim = nvim.with_hook(DecodeHook(encoding))
        return nvim



import os
from nose.tools import with_setup, eq_ as eq, ok_ as ok
from common import vim, cleanup


@with_setup(setup=cleanup)
def test_buffer():
    eq(vim.current.buffer, vim.windows[0].buffer)
    vim.command('new')
    vim.current.window = vim.windows[1]
    eq(vim.current.buffer, vim.windows[1].buffer)
    ok(vim.windows[0].buffer != vim.windows[1].buffer)


@with_setup(setup=cleanup)
def test_cursor():
    eq(vim.current.window.cursor, [1, 0])
    vim.command('normal ityping\033o  some text')
    eq(vim.current.buffer[:], ['typing', '  some text'])
    eq(vim.current.window.cursor, [2, 10])
    vim.current.window.cursor = [2, 6]
    vim.command('normal i dumb')
    eq(vim.current.buffer[:], ['typing', '  some dumb text'])


@with_setup(setup=cleanup)
def test_height():
    vim.command('vsplit')
    eq(vim.windows[1].height, vim.windows[0].height)
    vim.current.window = vim.windows[1]
    vim.command('split')
    eq(vim.windows[1].height, vim.windows[0].height // 2)
    vim.windows[1].height = 2
    eq(vim.windows[1].height, 2)


@with_setup(setup=cleanup)
def test_width():
    vim.command('split')
    eq(vim.windows[1].width, vim.windows[0].width)
    vim.current.window = vim.windows[1]
    vim.command('vsplit')
    eq(vim.windows[1].width, vim.windows[0].width // 2)
    vim.windows[1].width = 2
    eq(vim.windows[1].width, 2)


@with_setup(setup=cleanup)
def test_vars():
    vim.current.window.vars['python'] = [1, 2, {'3': 1}]
    eq(vim.current.window.vars['python'], [1, 2, {'3': 1}])
    eq(vim.eval('w:python'), [1, 2, {'3': 1}])


@with_setup(setup=cleanup)
def test_options():
    vim.current.window.options['colorcolumn'] = '4,3'
    eq(vim.current.window.options['colorcolumn'], '4,3')
    # global-local option
    vim.current.window.options['statusline'] = 'window-status'
    eq(vim.current.window.options['statusline'], 'window-status')
    eq(vim.options['statusline'], '')


@with_setup(setup=cleanup)
def test_position():
    height = vim.windows[0].height
    width = vim.windows[0].width
    vim.command('split')
    vim.command('vsplit')
    eq((vim.windows[0].row, vim.windows[0].col), (0, 0))
    vsplit_pos = width / 2
    split_pos = height / 2
    eq(vim.windows[1].row, 0)
    ok(vsplit_pos - 1 <= vim.windows[1].col <= vsplit_pos + 1)
    ok(split_pos - 1 <= vim.windows[2].row <= split_pos + 1)
    eq(vim.windows[2].col, 0)


@with_setup(setup=cleanup)
def test_tabpage():
    vim.command('tabnew')
    vim.command('vsplit')
    eq(vim.windows[0].tabpage, vim.tabpages[0])
    eq(vim.windows[1].tabpage, vim.tabpages[1])
    eq(vim.windows[2].tabpage, vim.tabpages[1])


@with_setup(setup=cleanup)
def test_valid():
    vim.command('split')
    window = vim.windows[1]
    vim.current.window = window
    ok(window.valid)
    vim.command('q')
    ok(not window.valid)



"""API for working with Nvim windows."""
from .common import Remote, RemoteMap


__all__ = ('Window')


class Window(Remote):

    """A remote Nvim window."""

    def __init__(self, session, code_data):
        """Initialize from session and code_data immutable object.

        The `code_data` contains serialization information required for
        msgpack-rpc calls. It must be immutable for Window equality to work.
        """
        self._session = session
        self.code_data = code_data
        self.vars = RemoteMap(session, 'window_get_var', 'window_set_var',
                              self)
        self.options = RemoteMap(session, 'window_get_option',
                                 'window_set_option', self)

    @property
    def buffer(self):
        """Get the `Buffer` currently being displayed by the window."""
        return self._session.request('window_get_buffer', self)

    @property
    def cursor(self):
        """Get the (row, col) tuple with the current cursor position."""
        return self._session.request('window_get_cursor', self)

    @cursor.setter
    def cursor(self, pos):
        """Set the (row, col) tuple as the new cursor position."""
        return self._session.request('window_set_cursor', self, pos)

    @property
    def height(self):
        """Get the window height in rows."""
        return self._session.request('window_get_height', self)

    @height.setter
    def height(self, height):
        """Set the window height in rows."""
        return self._session.request('window_set_height', self, height)

    @property
    def width(self):
        """Get the window width in rows."""
        return self._session.request('window_get_width', self)

    @width.setter
    def width(self, width):
        """Set the window height in rows."""
        return self._session.request('window_set_width', self, width)

    @property
    def row(self):
        """0-indexed, on-screen window position(row) in display cells."""
        return self._session.request('window_get_position', self)[0]

    @property
    def col(self):
        """0-indexed, on-screen window position(col) in display cells."""
        return self._session.request('window_get_position', self)[1]

    @property
    def tabpage(self):
        """Get the `Tabpage` that contains the window."""
        return self._session.request('window_get_tabpage', self)

    @property
    def valid(self):
        """Return True if the window still exists."""
        return self._session.request('window_is_valid', self)



"""Code for supporting compatibility across python versions."""

import sys
from imp import find_module as original_find_module


IS_PYTHON3 = sys.version_info >= (3, 0)


if IS_PYTHON3:
    def find_module(fullname, path):
        """Compatibility wrapper for imp.find_module.

        Automatically decodes arguments of find_module, in Python3
        they must be Unicode
        """
        if isinstance(fullname, bytes):
            fullname = fullname.decode()
        if isinstance(path, bytes):
            path = path.decode()
        elif isinstance(path, list):
            newpath = []
            for element in path:
                if isinstance(element, bytes):
                    newpath.append(element.decode())
                else:
                    newpath.append(element)
            path = newpath
        return original_find_module(fullname, path)

    # There is no 'long' type in Python3 just int
    long = int
else:
    find_module = original_find_module

NUM_TYPES = (int, long, float)



"""Event loop implementation that uses the `asyncio` standard module.

The `asyncio` module was added to python standard library on 3.4, and it
provides a pure python implementation of an event loop library. It is used
as a fallback in case pyuv is not available(on python implementations other
than CPython).

Earlier python versions are supported through the `trollius` package, which
is a backport of `asyncio` that works on Python 2.6+.
"""
from __future__ import absolute_import

import os
import sys
from collections import deque

try:
    # For python 3.4+, use the standard library module
    import asyncio
except ImportError:
    # Fallback to trollius
    import trollius as asyncio

from .base import BaseEventLoop


loop_cls = asyncio.SelectorEventLoop
if os.name == 'nt':
    # On windows use ProactorEventLoop which support pipes and is backed by the
    # more powerful IOCP facility
    loop_cls = asyncio.ProactorEventLoop


class AsyncioEventLoop(BaseEventLoop, asyncio.Protocol,
                       asyncio.SubprocessProtocol):

    """`BaseEventLoop` subclass that uses `asyncio` as a backend."""

    def connection_made(self, transport):
        """Used to signal `asyncio.Protocol` of a successful connection."""
        self._transport = transport
        if isinstance(transport, asyncio.SubprocessTransport):
            self._transport = transport.get_pipe_transport(0)

    def connection_lost(self, exc):
        """Used to signal `asyncio.Protocol` of a lost connection."""
        self._on_error(exc.message if exc else 'EOF')

    def data_received(self, data):
        """Used to signal `asyncio.Protocol` of incoming data."""
        if self._on_data:
            self._on_data(data)
            return
        self._queued_data.append(data)

    def pipe_connection_lost(self, fd, exc):
        """Used to signal `asyncio.SubprocessProtocol` of a lost connection."""
        self._on_error(exc.message if exc else 'EOF')

    def pipe_data_received(self, fd, data):
        """Used to signal `asyncio.SubprocessProtocol` of incoming data."""
        if fd == sys.stderr.fileno():
            self._on_stderr(data)
        elif self._on_data:
            self._on_data(data)
        else:
            self._queued_data.append(data)

    def process_exited(self):
        """Used to signal `asyncio.SubprocessProtocol` when the child exits."""
        self._on_error('EOF')

    def _init(self):
        self._loop = loop_cls()
        self._queued_data = deque()
        self._fact = lambda: self

    def _connect_tcp(self, address, port):
        coroutine = self._loop.create_connection(self._fact, address, port)
        self._loop.run_until_complete(coroutine)

    def _connect_socket(self, path):
        coroutine = self._loop.create_unix_connection(self._fact, path)
        self._loop.run_until_complete(coroutine)

    def _connect_stdio(self):
        coroutine = self._loop.connect_read_pipe(self._fact, sys.stdin)
        self._loop.run_until_complete(coroutine)
        coroutine = self._loop.connect_write_pipe(self._fact, sys.stdout)
        self._loop.run_until_complete(coroutine)

    def _connect_child(self, argv):
        coroutine = self._loop.subprocess_exec(self._fact, *argv)
        self._loop.run_until_complete(coroutine)

    def _start_reading(self):
        pass

    def _send(self, data):
        self._transport.write(data)

    def _run(self):
        while self._queued_data:
            self._on_data(self._queued_data.popleft())
        self._loop.run_forever()

    def _stop(self):
        self._loop.stop()

    def _threadsafe_call(self, fn):
        self._loop.call_soon_threadsafe(fn)

    def _setup_signals(self, signals):
        self._signals = list(signals)
        for signum in self._signals:
            self._loop.add_signal_handler(signum, self._on_signal, signum)

    def _teardown_signals(self):
        for signum in self._signals:
            self._loop.remove_signal_handler(signum)



"""Decorators used by python host plugin system."""
import inspect


import logging
logger = logging.getLogger(__name__)
debug, info, warn = (logger.debug, logger.info, logger.warn,)
__all__ = ('plugin', 'rpc_export', 'command', 'autocmd', 'function',
           'encoding', 'shutdown_hook')


def plugin(cls):
    """Tag a class as a plugin.

    This decorator is required to make the a class methods discoverable by the
    plugin_load method of the host.
    """
    cls._nvim_plugin = True
    # the _nvim_bind attribute is set to True by default, meaning that
    # decorated functions have a bound Nvim instance as first argument.
    # For methods in a plugin-decorated class this is not required, because
    # the class initializer will already receive the nvim object.
    for _, fn in inspect.getmembers(cls, inspect.ismethod):
        if hasattr(fn, '_nvim_bind'):
            fn.im_func._nvim_bind = False
    return cls


def rpc_export(rpc_method_name, sync=False):
    """Export a function or plugin method as a msgpack-rpc request handler."""
    def dec(f):
        f._nvim_rpc_method_name = rpc_method_name
        f._nvim_rpc_sync = sync
        f._nvim_bind = True
        f._nvim_prefix_plugin_path = False
        return f
    return dec


def command(name, nargs=0, complete=None, range=None, count=None, bang=False,
            register=False, sync=False, eval=None):
    """Tag a function or plugin method as a Nvim command handler."""
    def dec(f):
        f._nvim_rpc_method_name = 'command:{0}'.format(name)
        f._nvim_rpc_sync = sync
        f._nvim_bind = True
        f._nvim_prefix_plugin_path = True

        opts = {}

        if range:
            opts['range'] = '' if range is True else str(range)
        elif count:
            opts['count'] = count

        if bang:
            opts['bang'] = True

        if register:
            opts['register'] = True

        if nargs:
            opts['nargs'] = nargs

        if complete:
            opts['complete'] = complete

        if eval:
            opts['eval'] = eval

        f.nvim_rpc_spec = {
            'type': 'command',
            'name': name,
            'sync': sync,
            'opts': opts
        }
        return f
    return dec


def autocmd(name, pattern='*', sync=False, eval=None):
    """Tag a function or plugin method as a Nvim autocommand handler."""
    def dec(f):
        f._nvim_rpc_method_name = 'autocmd:{0}:{1}'.format(name, pattern)
        f._nvim_rpc_sync = sync
        f._nvim_bind = True
        f._nvim_prefix_plugin_path = True

        opts = {
            'pattern': pattern
        }

        if eval:
            opts['eval'] = eval

        f.nvim_rpc_spec = {
            'type': 'autocmd',
            'name': name,
            'sync': sync,
            'opts': opts
        }
        return f
    return dec


def function(name, range=False, sync=False, eval=None):
    """Tag a function or plugin method as a Nvim function handler."""
    def dec(f):
        f._nvim_rpc_method_name = 'function:{0}'.format(name)
        f._nvim_rpc_sync = sync
        f._nvim_bind = True
        f._nvim_prefix_plugin_path = True

        opts = {}

        if range:
            opts['range'] = '' if range is True else str(range)

        if eval:
            opts['eval'] = eval

        f.nvim_rpc_spec = {
            'type': 'function',
            'name': name,
            'sync': sync,
            'opts': opts
        }
        return f
    return dec


def shutdown_hook(f):
    """Tag a function or method as a shutdown hook."""
    f._nvim_shutdown_hook = True
    f._nvim_bind = True
    return f


def encoding(encoding=True):
    """Configure automatic encoding/decoding of strings."""
    def dec(f):
        f._nvim_encoding = encoding
        return f
    return dec



from nose.tools import with_setup, eq_ as eq
from common import vim, cleanup
from threading import Timer


@with_setup(setup=cleanup)
def test_interrupt_from_another_thread():
    session = vim.session
    timer = Timer(0.5, lambda: session.threadsafe_call(lambda: session.stop()))
    timer.start()
    eq(vim.session.next_message(), None)



"""Common code for graphical and text UIs."""
from ..compat import IS_PYTHON3


__all__ = ('Screen',)


if not IS_PYTHON3:
    range = xrange  # NOQA


class Cell(object):
    def __init__(self):
        self.text = ' '
        self.attrs = None

    def __repr__(self):
        return self.text

    def get(self):
        return self.text, self.attrs

    def set(self, text, attrs):
        self.text = text
        self.attrs = attrs

    def copy(self, other):
        other.text = self.text
        other.attrs = self.attrs


class Screen(object):

    """Store nvim screen state."""

    def __init__(self, columns, rows):
        """Initialize the Screen instance."""
        self.columns = columns
        self.rows = rows
        self.row = 0
        self.col = 0
        self.top = 0
        self.bot = rows - 1
        self.left = 0
        self.right = columns - 1
        self._cells = [[Cell() for c in range(columns)] for r in range(rows)]

    def clear(self):
        """Clear the screen."""
        self._clear_region(self.top, self.bot, self.left, self.right)

    def eol_clear(self):
        """Clear from the cursor position to the end of the scroll region."""
        self._clear_region(self.row, self.row, self.col, self.right)

    def cursor_goto(self, row, col):
        """Change the virtual cursor position."""
        self.row = row
        self.col = col

    def set_scroll_region(self, top, bot, left, right):
        """Change scroll region."""
        self.top = top
        self.bot = bot
        self.left = left
        self.right = right

    def scroll(self, count):
        """Shift scroll region."""
        top, bot = self.top, self.bot
        left, right = self.left, self.right
        if count > 0:
            start = top
            stop = bot - count + 1
            step = 1
        else:
            start = bot
            stop = top - count - 1
            step = -1
        # shift the cells
        for row in range(start, stop, step):
            target_row = self._cells[row]
            source_row = self._cells[row + count]
            for col in range(left, right + 1):
                tgt = target_row[col]
                source_row[col].copy(tgt)
        # clear invalid cells
        for row in range(stop, stop + count, step):
            self._clear_region(row, row, left, right)

    def put(self, text, attrs):
        """Put character on virtual cursor position."""
        cell = self._cells[self.row][self.col]
        cell.set(text, attrs)
        self.cursor_goto(self.row, self.col + 1)

    def get_cell(self, row, col):
        """Get text, attrs at row, col."""
        return self._cells[row][col].get()

    def get_cursor(self):
        """Get text, attrs at the virtual cursor position."""
        return self.get_cell(self.row, self.col)

    def iter(self, startrow, endrow, startcol, endcol):
        """Extract text/attrs at row, startcol-endcol."""
        for row in range(startrow, endrow + 1):
            r = self._cells[row]
            cell = r[startcol]
            curcol = startcol
            attrs = cell.attrs
            buf = [cell.text]
            for col in range(startcol + 1, endcol + 1):
                cell = r[col]
                if cell.attrs != attrs or not cell.text:
                    yield row, curcol, ''.join(buf), attrs
                    attrs = cell.attrs
                    buf = [cell.text]
                    curcol = col
                    if not cell.text:
                        # glyph uses two cells, yield a separate entry
                        yield row, curcol, '', None
                        curcol += 1
                else:
                    buf.append(cell.text)
            if buf:
                yield row, curcol, ''.join(buf), attrs

    def _clear_region(self, top, bot, left, right):
        for rownum in range(top, bot + 1):
            row = self._cells[rownum]
            for colnum in range(left, right + 1):
                cell = row[colnum]
                cell.set(' ', None)



"""Main Nvim interface."""
import os

from msgpack import ExtType

from .buffer import Buffer
from .common import (DecodeHook, Remote, RemoteMap, RemoteSequence,
                     SessionFilter, SessionHook, walk)
from .tabpage import Tabpage
from .window import Window
from ..compat import IS_PYTHON3


__all__ = ('Nvim')


os_chdir = os.chdir
os_fchdir = os.fchdir


class Nvim(object):

    """Class that represents a remote Nvim instance.

    This class is main entry point to Nvim remote API, it is a thin wrapper
    around Session instances.

    The constructor of this class must not be called directly. Instead, the
    `from_session` class method should be used to create the first instance
    from a raw `Session` instance.

    Subsequent instances for the same session can be created by calling the
    `with_hook` instance method and passing a SessionHook instance. This can
    be useful to have multiple `Nvim` objects that behave differently without
    one affecting the other.
    """

    @classmethod
    def from_session(cls, session):
        """Create a new Nvim instance for a Session instance.

        This method must be called to create the first Nvim instance, since it
        queries Nvim metadata for type information and sets a SessionHook for
        creating specialized objects from Nvim remote handles.
        """
        session.error_wrapper = lambda e: NvimError(e[1])
        channel_id, metadata = session.request(b'vim_get_api_info')

        encoding = session.request(b'vim_get_option', b'encoding')
        session._async_session._msgpack_stream.set_packer_encoding(encoding)

        if IS_PYTHON3:
            hook = DecodeHook()
            # decode all metadata strings for python3
            metadata = walk(hook.from_nvim, metadata, None, None, None)

        types = {
            metadata['types']['Buffer']['id']: Buffer,
            metadata['types']['Window']['id']: Window,
            metadata['types']['Tabpage']['id']: Tabpage,
        }

        return cls(session, channel_id, metadata).with_hook(ExtHook(types))

    def __init__(self, session, channel_id, metadata):
        """Initialize a new Nvim instance. This method is module-private."""
        self._session = session
        self.channel_id = channel_id
        self.metadata = metadata
        self.vars = RemoteMap(session, 'vim_get_var', 'vim_set_var')
        self.vvars = RemoteMap(session, 'vim_get_vvar', None)
        self.options = RemoteMap(session, 'vim_get_option', 'vim_set_option')
        self.buffers = RemoteSequence(session, 'vim_get_buffers')
        self.windows = RemoteSequence(session, 'vim_get_windows')
        self.tabpages = RemoteSequence(session, 'vim_get_tabpages')
        self.current = Current(session)
        self.error = NvimError

    def with_hook(self, hook):
        """Initialize a new Nvim instance."""
        return Nvim(SessionFilter(self.session, hook), self.channel_id,
                    self.metadata)

    @property
    def session(self):
        """Return the Session or SessionFilter for a Nvim instance."""
        return self._session

    def ui_attach(self, width, height, rgb):
        """Register as a remote UI.

        After this method is called, the client will receive redraw
        notifications.
        """
        return self._session.request('ui_attach', width, height, rgb)

    def ui_detach(self):
        """Unregister as a remote UI."""
        return self._session.request('ui_detach')

    def ui_try_resize(self, width, height):
        """Notify nvim that the client window has resized.

        If possible, nvim will send a redraw request to resize.
        """
        return self._session.request('ui_try_resize', width, height)

    def subscribe(self, event):
        """Subscribe to a Nvim event."""
        return self._session.request('vim_subscribe', event)

    def unsubscribe(self, event):
        """Unsubscribe to a Nvim event."""
        return self._session.request('vim_unsubscribe', event)

    def command(self, string):
        """Execute a single ex command."""
        return self._session.request('vim_command', string)

    def eval(self, string):
        """Evaluate a vimscript expression."""
        return self._session.request('vim_eval', string)

    def strwidth(self, string):
        """Return the number of display cells `string` occupies.

        Tab is counted as one cell.
        """
        return self._session.request('vim_strwidth', string)

    def list_runtime_paths(self):
        """Return a list of paths contained in the 'runtimepath' option."""
        return self._session.request('vim_list_runtime_paths')

    def foreach_rtp(self, cb):
        """Invoke `cb` for each path in 'runtimepath'.

        Call the given callable for each path in 'runtimepath' until either
        callable returns something but None, the exception is raised or there
        are no longer paths. If stopped in case callable returned non-None,
        vim.foreach_rtp function returns the value returned by callable.
        """
        for path in self._session.request('vim_list_runtime_paths'):
            try:
                if cb(path) is not None:
                    break
            except Exception:
                break

    def chdir(self, dir_path):
        """Run os.chdir, then all appropriate vim stuff."""
        os_chdir(dir_path)
        return self._session.request('vim_change_directory', dir_path)

    def fchdir(self, dir_fd):
        """Run os.chdir, then all appropriate vim stuff."""
        os_fchdir(dir_fd)
        return self._session.request('vim_change_directory', os.getcwd())

    def feedkeys(self, keys, options='', escape_csi=True):
        """Push `keys` to Nvim user input buffer.

        Options can be a string with the following character flags:
        - 'm': Remap keys. This is default.
        - 'n': Do not remap keys.
        - 't': Handle keys as if typed; otherwise they are handled as if coming
               from a mapping. This matters for undo, opening folds, etc.
        """
        return self._session.request('vim_feedkeys', keys, options, escape_csi)

    def input(self, bytes):
        """Push `bytes` to Nvim low level input buffer.

        Unlike `feedkeys()`, this uses the lowest level input buffer and the
        call is not deferred. It returns the number of bytes actually
        written(which can be less than what was requested if the buffer is
        full).
        """
        return self._session.request('vim_input', bytes)

    def replace_termcodes(self, string, from_part=False, do_lt=True,
                          special=True):
        r"""Replace any terminal code strings by byte sequences.

        The returned sequences are Nvim's internal representation of keys,
        for example:

        <esc> -> '\x1b'
        <cr>  -> '\r'
        <c-l> -> '\x0c'
        <up>  -> '\x80ku'

        The returned sequences can be used as input to `feedkeys`.
        """
        return self._session.request('vim_replace_termcodes', string,
                                     from_part, do_lt, special)

    def out_write(self, msg):
        """Print `msg` as a normal message."""
        return self._session.request('vim_out_write', msg)

    def err_write(self, msg):
        """Print `msg` as an error message."""
        return self._session.request('vim_err_write', msg)

    def quit(self, quit_command='qa!'):
        """Send a quit command to Nvim.

        By default, the quit command is 'qa!' which will make Nvim quit without
        saving anything.
        """
        try:
            self.command(quit_command)
        except IOError:
            # sending a quit command will raise an IOError because the
            # connection is closed before a response is received. Safe to
            # ignore it.
            pass


class Current(object):

    """Helper class for emulating vim.current from python-vim."""

    def __init__(self, session):
        self._session = session
        self.range = None

    @property
    def line(self):
        return self._session.request('vim_get_current_line')

    @line.setter
    def line(self, line):
        return self._session.request('vim_set_current_line', line)

    @property
    def buffer(self):
        return self._session.request('vim_get_current_buffer')

    @buffer.setter
    def buffer(self, buffer):
        return self._session.request('vim_set_current_buffer', buffer)

    @property
    def window(self):
        return self._session.request('vim_get_current_window')

    @window.setter
    def window(self, window):
        return self._session.request('vim_set_current_window', window)

    @property
    def tabpage(self):
        return self._session.request('vim_get_current_tabpage')

    @tabpage.setter
    def tabpage(self, tabpage):
        return self._session.request('vim_set_current_tabpage', tabpage)


class ExtHook(SessionHook):
    def __init__(self, types):
        self.types = types
        super(ExtHook, self).__init__(from_nvim=self.from_ext,
                                      to_nvim=self.to_ext)

    def from_ext(self, obj, session, method, kind):
        if type(obj) is ExtType:
            cls = self.types[obj.code]
            return cls(session, (obj.code, obj.data))
        return obj

    def to_ext(self, obj, session, method, kind):
        if isinstance(obj, Remote):
            return ExtType(*obj.code_data)
        return obj


class NvimError(Exception):
    pass



"""Event loop abstraction subpackage.

Tries to use pyuv as a backend, falling back to the asyncio implementation.
"""
try:
    # libuv is fully implemented in C, use it when available
    from .uv import UvEventLoop
    EventLoop = UvEventLoop
except ImportError:
    # asyncio(trollius on python 2) is pure python and should be more portable
    # across python implementations
    from .asyncio import AsyncioEventLoop
    EventLoop = AsyncioEventLoop


__all__ = ('EventLoop')



import platform
import sys

from setuptools import setup

install_requires = [
    'click>=3.0',
    'msgpack-python>=0.4.0',
]

if sys.version_info < (3, 4):
    # trollius is just a backport of 3.4 asyncio module
    install_requires.append('trollius')

has_cython = False
if not platform.python_implementation() == 'PyPy':
    # pypy already includes an implementation of the greenlet module
    install_requires.append('greenlet')
    try:
        # Cythonizing screen.py to improve scrolling/clearing speed. Maybe the
        # performance can be improved even further by writing a screen.pxd with
        # static type information
        from Cython.Build import cythonize
        has_cython = True
    except ImportError:
        pass

setup(name='neovim',
      version='0.0.27',
      description='Python client to neovim',
      url='http://github.com/neovim/python-client',
      download_url='https://github.com/neovim/python-client/archive/0.0.27.tar.gz',
      author='Thiago de Arruda',
      author_email='tpadilha84@gmail.com',
      license='MIT',
      packages=['neovim', 'neovim.api', 'neovim.msgpack_rpc', 'neovim.ui',
                'neovim.msgpack_rpc.event_loop', 'neovim.plugin'],
      install_requires=install_requires,
      ext_modules=cythonize('neovim/ui/screen.py') if has_cython else None,
      entry_points='''
      [console_scripts]
      pynvim=neovim.ui.cli:main
      ''',
      zip_safe=False)



"""Neovim Gtk+ UI."""
import math

import cairo

from gi.repository import GLib, GObject, Gdk, Gtk, Pango, PangoCairo

from .screen import Screen


__all__ = ('GtkUI',)


SHIFT = Gdk.ModifierType.SHIFT_MASK
CTRL = Gdk.ModifierType.CONTROL_MASK
ALT = Gdk.ModifierType.MOD1_MASK


# Translation table for the names returned by Gdk.keyval_name that don't match
# the corresponding nvim key names.
KEY_TABLE = {
    'BackSpace': 'BS',
    'Return': 'CR',
    'Escape': 'Esc',
    'Delete': 'Del',
    'Page_Up': 'PageUp',
    'Page_Down': 'PageDown',
    'Enter': 'CR',
    'ISO_Left_Tab': 'Tab'
}


if (GLib.MAJOR_VERSION, GLib.MINOR_VERSION,) <= (2, 32,):
    GLib.threads_init()


class GtkUI(object):

    """Gtk+ UI class."""

    def __init__(self):
        """Initialize the UI instance."""
        self._redraw_arg = None
        self._foreground = -1
        self._background = -1
        self._font_size = 13
        self._font_name = 'Monospace'
        self._screen = None
        self._attrs = None
        self._cursor_enabled = False
        self._mouse_enabled = False
        self._insert_cursor = False
        self._blink = False
        self._blink_timer_id = None
        self._resize_timer_id = None
        self._pressed = None
        self._invalid = None
        self._pending = [0, 0, 0]
        self._reset_cache()

    def start(self, bridge):
        """Start the UI event loop."""
        bridge.attach(80, 24, True)
        drawing_area = Gtk.DrawingArea()
        drawing_area.connect('draw', self._gtk_draw)
        window = Gtk.Window()
        window.add(drawing_area)
        window.set_events(window.get_events() |
                          Gdk.EventMask.BUTTON_PRESS_MASK |
                          Gdk.EventMask.BUTTON_RELEASE_MASK |
                          Gdk.EventMask.POINTER_MOTION_MASK |
                          Gdk.EventMask.SCROLL_MASK)
        window.connect('configure-event', self._gtk_configure)
        window.connect('delete-event', self._gtk_quit)
        window.connect('key-press-event', self._gtk_key)
        window.connect('button-press-event', self._gtk_button_press)
        window.connect('button-release-event', self._gtk_button_release)
        window.connect('motion-notify-event', self._gtk_motion_notify)
        window.connect('scroll-event', self._gtk_scroll)
        window.show_all()
        im_context = Gtk.IMContextSimple()
        im_context.connect('commit', self._gtk_input)
        self._pango_context = drawing_area.create_pango_context()
        self._drawing_area = drawing_area
        self._window = window
        self._im_context = im_context
        self._bridge = bridge
        Gtk.main()

    def quit(self):
        """Exit the UI event loop."""
        GObject.idle_add(Gtk.main_quit)

    def schedule_screen_update(self, apply_updates):
        """Schedule screen updates to run in the UI event loop."""
        def wrapper():
            apply_updates()
            self._flush()
            self._start_blinking()
            self._screen_invalid()
        GObject.idle_add(wrapper)

    def _screen_invalid(self):
        self._drawing_area.queue_draw()

    def _nvim_resize(self, columns, rows):
        da = self._drawing_area
        # create FontDescription object for the selected font/size
        font_str = '{0} {1}'.format(self._font_name, self._font_size)
        self._font, pixels, normal_width, bold_width = _parse_font(font_str)
        # calculate the letter_spacing required to make bold have the same
        # width as normal
        self._bold_spacing = normal_width - bold_width
        cell_pixel_width, cell_pixel_height = pixels
        # calculate the total pixel width/height of the drawing area
        pixel_width = cell_pixel_width * columns
        pixel_height = cell_pixel_height * rows
        gdkwin = da.get_window()
        content = cairo.CONTENT_COLOR
        self._cairo_surface = gdkwin.create_similar_surface(content,
                                                            pixel_width,
                                                            pixel_height)
        self._cairo_context = cairo.Context(self._cairo_surface)
        self._pango_layout = PangoCairo.create_layout(self._cairo_context)
        self._pango_layout.set_alignment(Pango.Alignment.LEFT)
        self._pango_layout.set_font_description(self._font)
        self._pixel_width, self._pixel_height = pixel_width, pixel_height
        self._cell_pixel_width = cell_pixel_width
        self._cell_pixel_height = cell_pixel_height
        self._screen = Screen(columns, rows)
        self._window.resize(pixel_width, pixel_height)

    def _nvim_clear(self):
        self._clear_region(self._screen.top, self._screen.bot + 1,
                           self._screen.left, self._screen.right + 1)
        self._screen.clear()

    def _nvim_eol_clear(self):
        row, col = self._screen.row, self._screen.col
        self._clear_region(row, row + 1, col, self._screen.right + 1)
        self._screen.eol_clear()

    def _nvim_cursor_goto(self, row, col):
        self._screen.cursor_goto(row, col)

    def _nvim_cursor_on(self):
        self._cursor_enabled = True

    def _nvim_cursor_off(self):
        self._cursor_enabled = False

    def _nvim_mouse_on(self):
        self._mouse_enabled = True

    def _nvim_mouse_off(self):
        self._mouse_enabled = False

    def _nvim_insert_mode(self):
        self._insert_cursor = True

    def _nvim_normal_mode(self):
        self._insert_cursor = False

    def _nvim_set_scroll_region(self, top, bot, left, right):
        self._screen.set_scroll_region(top, bot, left, right)

    def _nvim_scroll(self, count):
        self._flush()
        top, bot = self._screen.top, self._screen.bot + 1
        left, right = self._screen.left, self._screen.right + 1
        # The diagrams below illustrate what will happen, depending on the
        # scroll direction. "=" is used to represent the SR(scroll region)
        # boundaries and "-" the moved rectangles. note that dst and src share
        # a common region
        if count > 0:
            # move an rectangle in the SR up, this can happen while scrolling
            # down
            # +-------------------------+
            # | (clipped above SR)      |            ^
            # |=========================| dst_top    |
            # | dst (still in SR)       |            |
            # +-------------------------+ src_top    |
            # | src (moved up) and dst  |            |
            # |-------------------------| dst_bot    |
            # | src (cleared)           |            |
            # +=========================+ src_bot
            src_top, src_bot = top + count, bot
            dst_top, dst_bot = top, bot - count
            clr_top, clr_bot = dst_bot, src_bot
        else:
            # move a rectangle in the SR down, this can happen while scrolling
            # up
            # +=========================+ src_top
            # | src (cleared)           |            |
            # |------------------------ | dst_top    |
            # | src (moved down) and dst|            |
            # +-------------------------+ src_bot    |
            # | dst (still in SR)       |            |
            # |=========================| dst_bot    |
            # | (clipped below SR)      |            v
            # +-------------------------+
            src_top, src_bot = top, bot + count
            dst_top, dst_bot = top - count, bot
            clr_top, clr_bot = src_top, dst_top
        self._cairo_surface.flush()
        self._cairo_context.save()
        # The move is performed by setting the source surface to itself, but
        # with a coordinate transformation.
        _, y = self._get_coords(dst_top - src_top, 0)
        self._cairo_context.set_source_surface(self._cairo_surface, 0, y)
        # Clip to ensure only dst is affected by the change
        self._mask_region(dst_top, dst_bot, left, right)
        # Do the move
        self._cairo_context.paint()
        self._cairo_context.restore()
        # Clear the emptied region
        self._clear_region(clr_top, clr_bot, left, right)
        self._screen.scroll(count)

    def _nvim_highlight_set(self, attrs):
        self._attrs = self._get_pango_attrs(attrs)

    def _nvim_put(self, text):
        if self._screen.row != self._pending[0]:
            # flush pending text if jumped to a different row
            self._flush()
        # work around some redraw glitches that can happen
        self._redraw_glitch_fix()
        # Update internal screen
        self._screen.put(self._get_pango_text(text), self._attrs)
        self._pending[1] = min(self._screen.col - 1, self._pending[1])
        self._pending[2] = max(self._screen.col, self._pending[2])

    def _nvim_bell(self):
        self._window.get_window().beep()

    def _nvim_visual_bell(self):
        pass

    def _nvim_update_fg(self, fg):
        self._foreground = fg
        self._reset_cache()

    def _nvim_update_bg(self, bg):
        self._background = bg
        self._reset_cache()

    def _nvim_suspend(self):
        self._window.iconify()

    def _nvim_set_title(self, title):
        self._window.set_title(title)

    def _nvim_set_icon(self, icon):
        self._window.set_icon_name(icon)

    def _gtk_draw(self, wid, cr):
        if not self._screen:
            return
        # from random import random
        # cr.rectangle(0, 0, self._pixel_width, self._pixel_height)
        # cr.set_source_rgb(random(), random(), random())
        # cr.fill()
        self._cairo_surface.flush()
        cr.save()
        cr.rectangle(0, 0, self._pixel_width, self._pixel_height)
        cr.clip()
        cr.set_source_surface(self._cairo_surface, 0, 0)
        cr.paint()
        cr.restore()
        if self._cursor_enabled and self._blink:
            # Cursor is drawn separately in the window. This approach is
            # simpler because it doesn't taint the internal cairo surface,
            # which is used for scrolling
            row, col = self._screen.row, self._screen.col
            text, attrs = self._screen.get_cursor()
            self._pango_draw(row, col, [(text, attrs,)], cr=cr, cursor=True)

    def _gtk_configure(self, widget, event):
        def resize(*args):
            self._resize_timer_id = None
            width, height = self._window.get_size()
            columns = width / self._cell_pixel_width
            rows = height / self._cell_pixel_height
            if self._screen.columns == columns and self._screen.rows == rows:
                return
            self._bridge.resize(columns, rows)

        if not self._screen:
            return
        if event.width == self._pixel_width and \
           event.height == self._pixel_height:
            return
        if self._resize_timer_id is not None:
            GLib.source_remove(self._resize_timer_id)
        self._resize_timer_id = GLib.timeout_add(250, resize)

    def _gtk_quit(self, *args):
        self._bridge.exit()

    def _gtk_key(self, widget, event, *args):
        # This function was adapted from pangoterm source code
        keyval = event.keyval
        state = event.state
        # GtkIMContext will eat a Shift-Space and not tell us about shift.
        # Also don't let IME eat any GDK_KEY_KP_ events
        done = (False if state & SHIFT and keyval == ord(' ') else
                False if Gdk.KEY_KP_Space <= keyval <= Gdk.KEY_KP_Divide else
                self._im_context.filter_keypress(event))
        if done:
            # input method handled keypress
            return True
        if event.is_modifier:
            # We don't need to track the state of modifier bits
            return
        # translate keyval to nvim key
        key_name = Gdk.keyval_name(keyval)
        if key_name.startswith('KP_'):
            key_name = key_name[3:]
        input_str = _stringify_key(KEY_TABLE.get(key_name, key_name), state)
        self._bridge.input(input_str)

    def _gtk_button_press(self, widget, event, *args):
        if not self._mouse_enabled or event.type != Gdk.EventType.BUTTON_PRESS:
            return
        button = 'Left'
        if event.button == 2:
            button = 'Middle'
        elif event.button == 3:
            button = 'Right'
        col = int(math.floor(event.x / self._cell_pixel_width))
        row = int(math.floor(event.y / self._cell_pixel_height))
        input_str = _stringify_key(button + 'Mouse', event.state)
        input_str += '<{0},{1}>'.format(col, row)
        self._bridge.input(input_str)
        self._pressed = button

    def _gtk_button_release(self, widget, event, *args):
        self._pressed = None

    def _gtk_motion_notify(self, widget, event, *args):
        if not self._mouse_enabled or not self._pressed:
            return
        col = int(math.floor(event.x / self._cell_pixel_width))
        row = int(math.floor(event.y / self._cell_pixel_height))
        input_str = _stringify_key(self._pressed + 'Drag', event.state)
        input_str += '<{0},{1}>'.format(col, row)
        self._bridge.input(input_str)

    def _gtk_scroll(self, widget, event, *args):
        if not self._mouse_enabled:
            return
        col = int(math.floor(event.x / self._cell_pixel_width))
        row = int(math.floor(event.y / self._cell_pixel_height))
        if event.direction == Gdk.ScrollDirection.UP:
            key = 'ScrollWheelUp'
        elif event.direction == Gdk.ScrollDirection.DOWN:
            key = 'ScrollWheelDown'
        else:
            return
        input_str = _stringify_key(key, event.state)
        input_str += '<{0},{1}>'.format(col, row)
        self._bridge.input(input_str)

    def _gtk_input(self, widget, input_str, *args):
        self._bridge.input(input_str.replace('<', '<lt>'))

    def _start_blinking(self):
        def blink(*args):
            self._blink = not self._blink
            self._screen_invalid()
            self._blink_timer_id = GLib.timeout_add(500, blink)
        if self._blink_timer_id:
            GLib.source_remove(self._blink_timer_id)
        self._blink = False
        blink()

    def _clear_region(self, top, bot, left, right):
        self._flush()
        self._cairo_context.save()
        self._mask_region(top, bot, left, right)
        r, g, b = _split_color(self._background)
        r, g, b = r / 255.0, g / 255.0, b / 255.0
        self._cairo_context.set_source_rgb(r, g, b)
        self._cairo_context.paint()
        self._cairo_context.restore()

    def _mask_region(self, top, bot, left, right, cr=None):
        if not cr:
            cr = self._cairo_context
        x1, y1, x2, y2 = self._get_rect(top, bot, left, right)
        cr.rectangle(x1, y1, x2 - x1, y2 - y1)
        cr.clip()

    def _get_rect(self, top, bot, left, right):
        x1, y1 = self._get_coords(top, left)
        x2, y2 = self._get_coords(bot, right)
        return x1, y1, x2, y2

    def _get_coords(self, row, col):
        x = col * self._cell_pixel_width
        y = row * self._cell_pixel_height
        return x, y

    def _flush(self):
        row, startcol, endcol = self._pending
        self._pending[0] = self._screen.row
        self._pending[1] = self._screen.col
        self._pending[2] = self._screen.col
        if startcol == endcol:
            return
        self._cairo_context.save()
        ccol = startcol
        buf = []
        bold = False
        for _, col, text, attrs in self._screen.iter(row, row, startcol,
                                                     endcol - 1):
            newbold = attrs and 'bold' in attrs[0]
            if newbold != bold or not text:
                if buf:
                    self._pango_draw(row, ccol, buf)
                bold = newbold
                buf = [(text, attrs,)]
                ccol = col
            else:
                buf.append((text, attrs,))
        if buf:
            self._pango_draw(row, ccol, buf)
        self._cairo_context.restore()

    def _pango_draw(self, row, col, data, cr=None, cursor=False):
        markup = []
        for text, attrs in data:
            if not attrs:
                attrs = self._get_pango_attrs(None)
            attrs = attrs[1] if cursor else attrs[0]
            markup.append('<span {0}>{1}</span>'.format(attrs, text))
        markup = ''.join(markup)
        self._pango_layout.set_markup(markup, -1)
        # Draw the text
        if not cr:
            cr = self._cairo_context
        x, y = self._get_coords(row, col)
        if cursor and self._insert_cursor:
            cr.rectangle(x, y, self._cell_pixel_width / 4,
                         self._cell_pixel_height)
            cr.clip()
        cr.move_to(x, y)
        PangoCairo.update_layout(cr, self._pango_layout)
        PangoCairo.show_layout(cr, self._pango_layout)
        _, r = self._pango_layout.get_pixel_extents()

    def _get_pango_text(self, text):
        rv = self._pango_text_cache.get(text, None)
        if rv is None:
            rv = GLib.markup_escape_text(text or '')
            self._pango_text_cache[text] = rv
        return rv

    def _get_pango_attrs(self, attrs):
        key = tuple(sorted((k, v,) for k, v in (attrs or {}).items()))
        rv = self._pango_attrs_cache.get(key, None)
        if rv is None:
            fg = self._foreground if self._foreground != -1 else 0
            bg = self._background if self._background != -1 else 0xffffff
            n = {
                'foreground': _split_color(fg),
                'background': _split_color(bg),
            }
            if attrs:
                for k, v in attrs.items():
                    if k in ['foreground', 'background']:
                        n[k] = _split_color(v)
                    elif k == 'reverse':
                        n['foreground'] = _invert_color(*n['foreground'])
                        n['background'] = _invert_color(*n['background'])
                    elif k == 'italic':
                        n['font_style'] = 'italic'
                    elif k == 'bold':
                        n['font_weight'] = 'bold'
                        if self._bold_spacing:
                            n['letter_spacing'] = str(self._bold_spacing)
                    elif k == 'underline':
                        n['underline'] = 'single'
            c = dict(n)
            c['foreground'] = _invert_color(*_split_color(fg))
            c['background'] = _invert_color(*_split_color(bg))
            c['foreground'] = _stringify_color(*c['foreground'])
            c['background'] = _stringify_color(*c['background'])
            n['foreground'] = _stringify_color(*n['foreground'])
            n['background'] = _stringify_color(*n['background'])
            n = ' '.join(['{0}="{1}"'.format(k, v) for k, v in n.items()])
            c = ' '.join(['{0}="{1}"'.format(k, v) for k, v in c.items()])
            rv = (n, c,)
            self._pango_attrs_cache[key] = rv
        return rv

    def _reset_cache(self):
        self._pango_text_cache = {}
        self._pango_attrs_cache = {}

    def _redraw_glitch_fix(self):
        row, col = self._screen.row, self._screen.col
        text, attrs = self._screen.get_cursor()
        # when updating cells in italic or bold words, the result can become
        # messy(characters can be clipped or leave remains when removed). To
        # prevent that, always update non empty sequences of cells and the
        # surrounding space.
        # find the start of the sequence
        lcol = col - 1
        while lcol >= 0:
            text, _ = self._screen.get_cell(row, lcol)
            lcol -= 1
            if text == ' ':
                break
        self._pending[1] = min(lcol + 1, self._pending[1])
        # find the end of the sequence
        rcol = col + 1
        while rcol < self._screen.columns:
            text, _ = self._screen.get_cell(row, rcol)
            rcol += 1
            if text == ' ':
                break
        self._pending[2] = max(rcol, self._pending[2])


def _split_color(n):
    return ((n >> 16) & 0xff, (n >> 8) & 0xff, n & 0xff,)


def _invert_color(r, g, b):
    return (255 - r, 255 - g, 255 - b,)


def _stringify_color(r, g, b):
    return '#{0:0{1}x}'.format((r << 16) + (g << 8) + b, 6)


def _stringify_key(key, state):
    send = []
    if state & SHIFT:
        send.append('S')
    if state & CTRL:
        send.append('C')
    if state & ALT:
        send.append('A')
    send.append(key)
    return '<' + '-'.join(send) + '>'


def _parse_font(font, cr=None):
    if not cr:
        ims = cairo.ImageSurface(cairo.FORMAT_RGB24, 300, 300)
        cr = cairo.Context(ims)
    fd = Pango.font_description_from_string(font)
    layout = PangoCairo.create_layout(cr)
    layout.set_font_description(fd)
    layout.set_alignment(Pango.Alignment.LEFT)
    layout.set_markup('<span font_weight="bold">A</span>')
    bold_width, _ = layout.get_size()
    layout.set_markup('<span>A</span>')
    pixels = layout.get_pixel_size()
    normal_width, _ = layout.get_size()
    return fd, pixels, normal_width, bold_width



"""Neovim tickit UI."""
import os
import re
import signal
import sys

import cffi
import cffi.verifier

from .screen import Screen
from ..compat import IS_PYTHON3

try:
    # For python 3.4+, use the standard library module
    import asyncio
except ImportError:
    # Fallback to trollius
    import trollius as asyncio


loop_cls = asyncio.SelectorEventLoop
if os.name == 'nt':
    # On windows use ProactorEventLoop which support pipes and is backed by
    # the more powerful IOCP facility
    loop_cls = asyncio.ProactorEventLoop


__all__ = ('TickitUI',)


if not IS_PYTHON3:
    range = xrange  # NOQA


# Translation table for the key name strings passed by libtickit that don't
# match
KEY_TABLE = {
    'Backspace': 'BS',
    'Enter': 'CR',
    'Escape': 'Esc',
    'Delete': 'Del'
}

EMPTY_ATTRS = {}

API = '''
enum {
  TICKIT_MOD_SHIFT = 0x01,
  TICKIT_MOD_ALT   = 0x02,
  TICKIT_MOD_CTRL  = 0x04,
};

enum {
  TICKIT_MOUSEWHEEL_UP = 1,
  TICKIT_MOUSEWHEEL_DOWN,
};

typedef enum {
  TICKIT_PEN_FG,
  TICKIT_PEN_BG,
  TICKIT_PEN_BOLD,
  TICKIT_PEN_UNDER,
  TICKIT_PEN_ITALIC,
  TICKIT_PEN_REVERSE,
  TICKIT_PEN_STRIKE,
  TICKIT_PEN_ALTFONT,
  TICKIT_PEN_BLINK,
  TICKIT_N_PEN_ATTRS
} TickitPenAttr;

typedef enum {
  TICKIT_TERM_MOUSEMODE_OFF,
  TICKIT_TERM_MOUSEMODE_CLICK,
  TICKIT_TERM_MOUSEMODE_DRAG,
  TICKIT_TERM_MOUSEMODE_MOVE,
} TickitTermMouseMode;

typedef enum {
  TICKIT_TERMCTL_ALTSCREEN = 1,
  TICKIT_TERMCTL_CURSORVIS,
  TICKIT_TERMCTL_MOUSE,
  TICKIT_TERMCTL_CURSORBLINK,
  TICKIT_TERMCTL_CURSORSHAPE,
  TICKIT_TERMCTL_ICON_TEXT,
  TICKIT_TERMCTL_TITLE_TEXT,
  TICKIT_TERMCTL_ICONTITLE_TEXT,
  TICKIT_TERMCTL_KEYPAD_APP,
  TICKIT_TERMCTL_COLORS,
} TickitTermCtl;

typedef enum {
  TICKIT_TERM_CURSORSHAPE_BLOCK = 1,
  TICKIT_TERM_CURSORSHAPE_UNDER,
  TICKIT_TERM_CURSORSHAPE_LEFT_BAR,
} TickitTermCursorShape;

typedef enum {
  TICKIT_EV_RESIZE = 0x01,
  TICKIT_EV_KEY    = 0x02,
  TICKIT_EV_MOUSE  = 0x04,
  TICKIT_EV_CHANGE = 0x08,
  TICKIT_EV_UNBIND = 0x80000000,
} TickitEventType;

typedef enum {
  TICKIT_KEYEV_KEY = 1,
  TICKIT_KEYEV_TEXT,
} TickitKeyEventType;

typedef enum {
  TICKIT_MOUSEEV_PRESS = 1,
  TICKIT_MOUSEEV_DRAG,
  TICKIT_MOUSEEV_RELEASE,
  TICKIT_MOUSEEV_WHEEL,
} TickitMouseEventType;

typedef struct {
  int         lines, cols;
  int         type;
  const char *str;
  int         button;
  int         line, col;
  int         mod;
} TickitEvent;

struct timeval {
    long tv_sec;
    long tv_usec;
};

typedef struct TickitTerm_ TickitTerm;
typedef struct TickitPen_ TickitPen;

typedef void TickitTermEventFn(TickitTerm *tt, TickitEventType ev,
                               TickitEvent *args, void *data);

TickitTerm *tickit_term_new(void);
void tickit_term_destroy(TickitTerm *tt);
void tickit_term_await_started(TickitTerm *tt, const struct timeval *timeout);
TickitPen *tickit_pen_new(void);
void tickit_term_setpen(TickitTerm *tt, TickitPen *pen);
void tickit_term_set_input_fd(TickitTerm *tt, int fd);
void tickit_term_set_output_fd(TickitTerm *tt, int fd);
int tickit_term_setctl_int(TickitTerm *tt, TickitTermCtl ctl, int value);
int tickit_term_setctl_str(TickitTerm *tt, TickitTermCtl ctl,
                           const char *value);
void tickit_term_input_readable(TickitTerm *tt);
int tickit_term_input_check_timeout(TickitTerm *tt);
void tickit_term_set_output_buffer(TickitTerm *tt, size_t len);
void tickit_term_flush(TickitTerm *tt);
void tickit_pen_set_colour_attr(TickitPen *pen, TickitPenAttr attr, int val);
void tickit_pen_set_bool_attr(TickitPen *pen, TickitPenAttr attr, bool val);
void tickit_term_printn(TickitTerm *tt, const char *str, size_t len);
int  tickit_term_goto(TickitTerm *tt, int line, int col);
void tickit_term_get_size(TickitTerm *tt, int *lines, int *cols);
void tickit_term_set_size(TickitTerm *tt, int lines, int cols);
void tickit_term_refresh_size(TickitTerm *tt);
void tickit_term_clear(TickitTerm *tt);
int tickit_term_bind_event(TickitTerm *tt, TickitEventType ev,
                           TickitTermEventFn *fn, void *user);
'''


class TickitVerifier(cffi.verifier.Verifier):
    def get_extension(self, *args, **kwargs):
        # Executed only the first time the native extension needs to be built
        from subprocess import check_output
        try:
            inc_dir = check_output(['pkg-config', '--cflags',
                                    'tickit']).strip()[2:]
            lib_dir = check_output(['pkg-config', '--libs', '--libs-only-L',
                                    'tickit']).strip()[2:]
            if os.path.isdir(inc_dir):
                self.kwds['include_dirs'] = [inc_dir]
            if os.path.isdir(lib_dir):
                link_flags = '-Wl,-rpath={0}'.format(lib_dir)
                self.kwds['library_dirs'] = [lib_dir]
                self.kwds['extra_link_args'] = [link_flags]
        except OSError as e:
            if e.errno != os.errno.ENOENT:
                raise
        return super(TickitVerifier, self).get_extension(*args, **kwargs)


ffi = cffi.FFI()
ffi.cdef(API)
lib = TickitVerifier(ffi, '''
#include <sys/time.h>
#include <tickit.h>
''', libraries=['tickit']).load_library()


class TickitUI(object):

    """Tickit UI class."""

    def start(self, bridge):
        """Start the UI event loop."""
        while True:
            self._suspended = False
            self._foreground = -1
            self._background = -1
            self._screen = None
            self._attrs = None
            self._printattrs = 0
            self._cursor_enabled = False
            self._invalid_rects = []  # regions that will be repainted on flush
            self._tickit_event_loop(bridge)
            if self._suspended:
                os.kill(0, signal.SIGTSTP)
            else:
                break

    def quit(self):
        """Exit the UI event loop."""
        def cb():
            self._loop.remove_reader(sys.stdin.fileno())
            self._loop.remove_signal_handler(signal.SIGINT)
            self._loop.remove_signal_handler(signal.SIGWINCH)
            self._loop.stop()
        self._loop.call_soon_threadsafe(cb)

    def schedule_screen_update(self, apply_updates):
        """Schedule screen updates to run in the UI event loop."""
        def cb():
            lib.tickit_term_setctl_int(self._tt,
                                       lib.TICKIT_TERMCTL_CURSORVIS, False)
            apply_updates()
            for top, bot, left, right in self._invalid_rects:
                self._repaint_rect(top, bot, left, right)
            del self._invalid_rects[:]
            if self._cursor_enabled:
                lib.tickit_term_setctl_int(self._tt,
                                           lib.TICKIT_TERMCTL_CURSORVIS, True)
            lib.tickit_term_flush(self._tt)

        self._loop.call_soon_threadsafe(cb)

    def _tickit_event(self, tt, ev_type, ev_args, *args):
        if ev_type == lib.TICKIT_EV_KEY:
            self._tickit_key(ev_args.str, ev_args.mod)

    def _tickit_key(self, type, str, mod):
        key_str = ffi.string(str)
        if type == lib.TICKIT_KEYEV_KEY:
            input_str = _stringify_key(KEY_TABLE.get(key_str, key_str), mod)
        else:
            input_str = key_str.replace('<', '<lt>')
        self._bridge.input(input_str)

    def _tickit_mouse(self, type, button, row, col, mod):
        if button == 1:
            key = 'Left'
        elif button == 2:
            key = 'Middle'
        elif button == 3:
            key = 'Right'
        else:
            return
        if type == lib.TICKIT_MOUSEEV_PRESS:
            key += 'Mouse'
        elif type == lib.TICKIT_MOUSEEV_DRAG:
            key += 'Drag'
        elif type == lib.TICKIT_MOUSEEV_WHEEL:
            if button == lib.TICKIT_MOUSEWHEEL_UP:
                key = 'ScrollWheelUp'
            else:
                key = 'ScrollWheelDown'
        input_str = _stringify_key(key, mod)
        input_str += '<{0},{1}>'.format(col, row)
        self._bridge.input(input_str)

    def _nvim_resize(self, columns, rows):
        lib.tickit_term_set_size(self._tt, rows, columns)
        self._screen = Screen(columns, rows)

    def _nvim_clear(self):
        self._screen.clear()
        self._repaint_rect(self._screen.top, self._screen.bot,
                           self._screen.left, self._screen.right)

    def _nvim_eol_clear(self):
        self._screen.eol_clear()
        self._repaint_rect(self._screen.row, self._screen.row,
                           self._screen.col, self._screen.right)

    def _nvim_cursor_goto(self, row, col):
        self._screen.cursor_goto(row, col)
        lib.tickit_term_goto(self._tt, row, col)

    def _nvim_cursor_on(self):
        self._cursor_enabled = True

    def _nvim_cursor_off(self):
        self._cursor_enabled = False

    def _nvim_mouse_on(self):
        lib.tickit_term_setctl_int(self._tt, lib.TICKIT_TERMCTL_MOUSE,
                                   lib.TICKIT_TERM_MOUSEMODE_DRAG)

    def _nvim_mouse_off(self):
        lib.tickit_term_setctl_int(self._tt, lib.TICKIT_TERMCTL_MOUSE,
                                   lib.TICKIT_TERM_MOUSEMODE_OFF)

    def _nvim_insert_mode(self):
        if lib.tickit_term_setctl_int(self._tt,
                                      lib.TICKIT_TERMCTL_CURSORSHAPE,
                                      lib.TICKIT_TERM_CURSORSHAPE_LEFT_BAR):
            return
        lib.tickit_term_setctl_int(self._tt, lib.TICKIT_TERMCTL_CURSORSHAPE,
                                   lib.TICKIT_TERM_CURSORSHAPE_UNDER)

    def _nvim_normal_mode(self):
        lib.tickit_term_setctl_int(self._tt, lib.TICKIT_TERMCTL_CURSORSHAPE,
                                   lib.TICKIT_TERM_CURSORSHAPE_BLOCK)

    def _nvim_set_scroll_region(self, top, bot, left, right):
        self._screen.set_scroll_region(top, bot, left, right)

    def _nvim_scroll(self, count):
        self._screen.scroll(count)
        self._invalidate(self._screen.top, self._screen.bot,
                         self._screen.left, self._screen.right)

    def _nvim_highlight_set(self, attrs):
        self._attrs = attrs

    def _nvim_put(self, text):
        self._screen.put(text, self._attrs)
        self._print(text, self._attrs)

    def _nvim_bell(self):
        lib.tickit_term_printn(self._tt, bytes('\x07'), 1)

    def _nvim_visual_bell(self):
        pass

    def _nvim_update_fg(self, fg):
        self._foreground = fg

    def _nvim_update_bg(self, bg):
        self._background = bg

    def _nvim_suspend(self):
        if not self._suspended:
            self.quit()
            self._suspended = True

    def _nvim_set_title(self, title):
        lib.tickit_term_setctl_str(self._tt, lib.TICKIT_TERMCTL_TITLE_TEXT,
                                   bytes(title))

    def _nvim_set_icon(self, icon):
        lib.tickit_term_setctl_str(self._tt, lib.TICKIT_TERMCTL_ICON_TEXT,
                                   bytes(icon))

    def _tickit_event_loop(self, bridge):
        width_ptr = ffi.new('int *')
        height_ptr = ffi.new('int *')
        tv = ffi.new('struct timeval *')
        tv.tv_usec = 50000
        tt = lib.tickit_term_new()
        lib.tickit_term_set_input_fd(tt, sys.stdin.fileno())
        lib.tickit_term_set_output_fd(tt, sys.stdout.fileno())
        lib.tickit_term_await_started(tt, tv)
        lib.tickit_term_setctl_int(tt, lib.TICKIT_TERMCTL_ALTSCREEN, 1)
        lib.tickit_term_refresh_size(tt)
        lib.tickit_term_get_size(tt, height_ptr, width_ptr)
        lib.tickit_term_clear(tt)

        def winch():
            lib.tickit_term_refresh_size(tt)
            lib.tickit_term_get_size(tt, height_ptr, width_ptr)
            bridge.resize(width_ptr[0], height_ptr[0])

        def interrupt():
            bridge.input('<C-c>')

        def input_check():
            check_timeout = lambda: lib.tickit_term_input_check_timeout(tt)
            lib.tickit_term_input_readable(tt)
            sleep = check_timeout()
            if sleep != -1:
                self._loop.call_later(sleep / 1000.0, check_timeout)

        @ffi.callback('TickitTermEventFn')
        def event(tt, ev_type, ev_args, *args):
            if ev_type == lib.TICKIT_EV_KEY:
                self._tickit_key(ev_args.type, ev_args.str, ev_args.mod)
            elif ev_type == lib.TICKIT_EV_MOUSE:
                self._tickit_mouse(ev_args.type, ev_args.button, ev_args.line,
                                   ev_args.col, ev_args.mod)

        pen = lib.tickit_pen_new()
        lib.tickit_term_setpen(tt, pen)
        lib.tickit_term_bind_event(tt, lib.TICKIT_EV_KEY | lib.TICKIT_EV_MOUSE,
                                   event, ffi.NULL)
        lib.tickit_term_set_output_buffer(tt, 0xffff)
        bridge.attach(width_ptr[0], height_ptr[0], False)
        self._bridge = bridge
        self._tt = tt
        self._pen = pen
        self._loop = loop_cls()
        self._loop.add_reader(sys.stdin.fileno(), input_check)
        self._loop.add_signal_handler(signal.SIGINT, interrupt)
        self._loop.add_signal_handler(signal.SIGWINCH, winch)
        self._loop.run_forever()
        lib.tickit_term_destroy(tt)
        bridge.detach()

    def _repaint_rect(self, top, bot, left, right):
        save_row, save_col = self._screen.row, self._screen.col
        for row, col, text, attrs in self._screen.iter(top, bot, left, right):
            lib.tickit_term_goto(self._tt, row, col)
            self._print(text, attrs)
        # restore cursor
        lib.tickit_term_goto(self._tt, save_row, save_col)

    def _print(self, text, attrs):
        if attrs != self._printattrs:
            if not attrs:
                attrs = EMPTY_ATTRS
            fg = attrs.get('foreground', self._foreground)
            bg = attrs.get('background', self._background)
            lib.tickit_pen_set_colour_attr(self._pen, lib.TICKIT_PEN_FG, fg)
            lib.tickit_pen_set_colour_attr(self._pen, lib.TICKIT_PEN_BG, bg)
            lib.tickit_pen_set_bool_attr(self._pen, lib.TICKIT_PEN_BOLD,
                                         'bold' in attrs)
            lib.tickit_pen_set_bool_attr(self._pen, lib.TICKIT_PEN_ITALIC,
                                         'italic' in attrs)
            lib.tickit_pen_set_bool_attr(self._pen, lib.TICKIT_PEN_UNDER,
                                         'underline' in attrs)
            lib.tickit_pen_set_bool_attr(self._pen, lib.TICKIT_PEN_REVERSE,
                                         'reverse' in attrs)
            lib.tickit_term_setpen(self._tt, self._pen)
            self._printattrs = attrs
        text = bytes(text)
        lib.tickit_term_printn(self._tt, text, len(text))

    def _invalidate(self, top, bot, left, right):
        # search the invalid rects for one that intersects the newly
        # invalidated rect
        intersects = None
        for i in range(len(self._invalid_rects)):
            t, b, l, r = self._invalid_rects[i]
            if top > b or bot < t or left > r or right < l:
                continue
            intersects = (t, b, l, r, i,)
            break
        if intersects:
            # replace the intersecting rect by the union with the newly
            # invalidated rect
            t, b, l, r, i = intersects
            top = min(top, t)
            bot = max(bot, b)
            left = min(left, l)
            right = max(right, r)
            self._invalid_rects[i] = (top, bot, left, right,)
        else:
            # add new entry
            self._invalid_rects.append((top, bot, left, right,))


MODIFIERS_AND_KEY = re.compile(r'^(?:(?:S|C|A|M)-)*(.+)$')


def _stringify_key(key, mod):
    key = MODIFIERS_AND_KEY.findall(key)[0]
    send = []
    if mod & lib.TICKIT_MOD_SHIFT:
        send.append('S')
    if mod & lib.TICKIT_MOD_CTRL:
        send.append('C')
    if mod & lib.TICKIT_MOD_ALT:
        send.append('A')
    send.append(key)
    return '<' + '-'.join(send) + '>'
